
// Parameterized implementation of List ADT
//   with representation classes hidden


object List {

  // creators

  def empty[A <: Ordered[A]] ():List[A] = new ListEmpty[A]()

  def singleton[B <: Ordered[B]] (i:B):List[B] = new ListSingleton[B](i)

  def merge[C <: Ordered[C]] (L:List[C], M:List[C]):List[C] = new ListMerge[C](L,M)

  
  private abstract class Common[T <: Ordered[T]] extends List[T] {

    def <= (M:List[T]):Boolean = 
      if (M.isEmpty())
	false
      else if (first()==M.first())
	rest() <= M.rest()
      else
	false
  }


  private class ListEmpty[T <: Ordered[T]] () extends Common[T] {

    def isEmpty ():Boolean = true
    def first ():T = throw new RuntimeException("empty().first()")
    def rest ():List[T] = throw new RuntimeException("empty().rest()")

    def length ():Int = 0

    override def hashCode ():Int = 41
    
    override def toString ():String = ""
  }



  private class ListSingleton[U <: Ordered[U]] (i:U) extends Common[U] {
    
    def isEmpty ():Boolean = false
    def first ():U = i
    def rest ():List[U] = List.empty()

    def length ():Int = 1

    override def hashCode ():Int = 41 + i.hashCode()

    override def toString ():String = " " + i.toString() 
  }


  
  private class ListMerge[V <: Ordered[V]] (L:List[V], M:List[V]) extends Common[V] {

    def isEmpty ():Boolean = 
      (L.isEmpty() && M.isEmpty())

    def first ():V = 
      if (L.isEmpty())
        M.first()
      else
        L.first()
    
    def rest ():List[V] = 
      if (L.isEmpty())
        M.rest()
      else
        List.merge(L.rest(),M)

    def length ():Int = L.length() + M.length()
    
    override def hashCode ():Int = 
      41 * (
        41 + L.hashCode()
      ) + M.hashCode()

    override def toString ():String = L.toString() + M.toString()
  }

}


abstract class List[T <: Ordered[T]] extends Ordered[List[T]] {

  def isEmpty ():Boolean
  def first ():T
  def rest ():List[T]
  def length ():Int
  def <= (M:List[T]):Boolean

}

